'''
Created on 14 Jan 2020

@author: si
'''
import copy
import csv
import os

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class CsvConnector(DataConnector):
    engine_type = 'csv://'
    optional_args = {'field_names': None}

    def __init__(self, *args, **kwargs):
        """
        Connector to Comma Separated Value (CSV) files.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for CsvConnector
         fieldnames (sequence, probably a list) - Field names to use for all rows in file. The
                             first line of file isn't the header with field names, it's data.
                             Note - this option may be merged into schemas (TODO).

        Connection information-
            engine_url format is
            csv://<filesystem absolute path>data_file.csv[;start=<line number>][;end=<line number>][;encoding=<character encoding>]
        e.g. csv:///data/my_project/all_the_data.csv
        """
        super().__init__(*args, **kwargs)

        # fieldnames are loaded from construction args or from field. This will be unified when
        # schemas are implemented. For now, keep track so loading fieldnames from file doesn't
        # make a :emthod:`_reset`
        self.base_field_names = copy.copy(self.field_names)

        self.delimiter = ','
        self._reset()

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError('Read+Write access not yet implemented')

    def _reset(self):
        self.file_handle = None
        self.csv = None
        self._encoding = None
        self._engine_params = None
        self.file_size = None
        self.approx_position = 0
        self.field_names = copy.copy(self.base_field_names)

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding .start and .end
        """
        if self._engine_params is None:
            self._engine_params = self.ignition._decode_filesystem_engine_url(
                self.engine_url,
                optional_args=['encoding', 'start', 'end']
            )

            if 'encoding' in self._engine_params:
                self._encoding = self.engine_params.encoding

            for typed_param in ['start', 'end']:
                if typed_param in self.engine_params:
                    self.engine_params[typed_param] = int(self.engine_params[typed_param])

            if 'start' in self._engine_params or 'end' in self._engine_params:
                raise NotImplementedError("TODO")

        return self._engine_params

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if 'encoding' in ep else 'utf-8-sig'
        return self._encoding

    def close_connection(self):
        if self.file_handle is not None:
            self.file_handle.close()
        self._reset()

    def connect(self):
        if self.csv is None:

            if self.access == AccessMode.READ:
                self.file_handle = open(self.engine_params.file_path, 'r', encoding=self.encoding)
                self.file_size = os.stat(self.engine_params.file_path).st_size

                self.csv = csv.DictReader(self.file_handle,
                                          delimiter=self.delimiter,
                                          fieldnames=self.field_names,
                                          )
                self.field_names = self.csv.fieldnames

            elif self.access == AccessMode.WRITE:

                # auto create directory
                file_dir = os.path.dirname(self.engine_params.file_path)
                if not os.path.exists(file_dir):
                    os.makedirs(file_dir)

                self.file_handle = open(self.engine_params.file_path,
                                        'w',
                                        newline='\n',
                                        encoding=self.encoding
                                        )
                self.csv = csv.DictWriter(self.file_handle,
                                          delimiter=self.delimiter,
                                          fieldnames=self.field_names,
                                          )
                self.csv.writeheader()

            else:
                raise ValueError('Unknown access mode')

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()
        for raw in self.csv:
            # OSError: telling position disabled by next() call so this for now
            # str(x) will slightly over count 'None'. None is given by DictReader when
            # trailing commas are omitted for optional fields at end of row.
            self.approx_position += len(self.delimiter.join([str(x) for x in raw.values()]))
            yield Pinnate(data=raw)

    @property
    def data(self):
        raise NotImplementedError("TODO")

    @property
    def schema(self):
        raise NotImplementedError("TODO")

    @property
    def progress(self):
        if self.access != AccessMode.READ or self.file_size is None or self.approx_position == 0:
            return None

        return self.approx_position / self.file_size

    def add(self, data):
        """
        Write line to CSV file.
        @param data: (dict or Pinnate)
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        # until schemas are implemented, first row determines fields
        if self.csv is None and self.field_names is None:
            if isinstance(data, dict):
                self.field_names = list(data.keys())
            elif isinstance(data, Pinnate):
                self.field_names = list(data.as_dict().keys())

        self.connect()

        if isinstance(data, dict):
            self.csv.writerow(data)
        elif isinstance(data, Pinnate):
            self.csv.writerow(data.as_dict())
        else:
            raise ValueError("data isn't an accepted type. Only (dict) or (Pinnate) are accepted.")


class TsvConnector(CsvConnector):
    """
    Tab separated values. See :class:`CsvConnector`
    """
    engine_type = 'tsv://'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # see note in CsvConnector
        self.base_field_names = copy.copy(self.field_names)
        self.delimiter = '\t'



import os
import tarfile
import tempfile
import unittest

import ayeaye
from ayeaye.connectors.flowerpot import FlowerpotEngine, FlowerPotConnector
from ayeaye.connectors.csv_connector import CsvConnector, TsvConnector
from ayeaye.connectors.json_connector import JsonConnector
from ayeaye.connectors.multi_connector import MultiConnector

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_FLOWERPOT_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'exampleflowerpot.tar.gz')
EXAMPLE_CSV_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'deadly_creatures.csv')
EXAMPLE_TSV_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'monkeys.tsv')
EXAMPLE_ENGINE_URL = 'gs+flowerpot://fake_flowerpot_bucket/some_file.json'
EXAMPLE_JSON_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'london_weather.json')
EXAMPLE_CSV_BROKEN_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'deadly_missing_values.csv')
EXAMPLE_CSV_MICE = os.path.join(PROJECT_TEST_PATH, 'data', 'mice.csv')
EXAMPLE_CSV_SQUIRRELS = os.path.join(PROJECT_TEST_PATH, 'data', 'squirrels.csv')


class TestConnectors(unittest.TestCase):

    def test_flowerpot_deserialize(self):
        test_string = bytes(
            '{"availability": "apple", "referential": "raspberry"}\n{"availability": "anchor", "referential": "rudder"}',
            encoding='utf-8')
        result = FlowerpotEngine._deserialize_ndjson_string(test_string)
        assert result[0] == {"availability": "apple", "referential": "raspberry"}
        assert result[1] == {"availability": "anchor", "referential": "rudder"}

    def test_iterate_over_json_lines(self):
        with tarfile.open(EXAMPLE_FLOWERPOT_PATH, 'r:gz') as tf:
            reader = FlowerpotEngine(tf)
            results = list(reader.items())
            assert len(results) == 4
            assert "availability" in results[0]
            assert "referential" in results[0]

    def test_flowerpot_all_items(self):
        """
        Iterate all the data items in all the files in the example flowerpot.
        """
        c = FlowerPotConnector(engine_url="flowerpot://" + EXAMPLE_FLOWERPOT_PATH)
        all_items = [(r.availability, r.referential) for r in c]
        all_items.sort()
        expected = "[('acoustic', 'rap'), ('anchor', 'rudder'), ('antenna', 'receive'), ('apple', 'raspberry')]"
        assert expected == str(all_items)

    def test_flowerpot_query_one_file(self):
        """
        The 'table' kwarg gets rows from all files that start with that string.
        """
        c = FlowerPotConnector(engine_url="flowerpot://" + EXAMPLE_FLOWERPOT_PATH)
        some_items = [(r.availability, r.referential) for r in c.query(table='test_a')]
        some_items.sort()
        expected = "[('anchor', 'rudder'), ('apple', 'raspberry')]"
        assert expected == str(some_items)

    def test_csv_basics(self):
        """
        Iterate all the data items and check each row is being yielded as an instance of
        :class:`ayeaye.Pinnate`
        """
        c = CsvConnector(engine_url="csv://" + EXAMPLE_CSV_PATH)

        animals_names = ", ".join([deadly_animal.common_name for deadly_animal in c])
        expected = 'Crown of thorns starfish, Golden dart frog'
        assert expected == animals_names

    def test_csv_write(self):
        """
        Write to a CSV without using a schema.
        """
        data_dir = tempfile.mkdtemp()
        csv_file = os.path.join(data_dir, "fish.csv")
        c = CsvConnector(engine_url="csv://" + csv_file, access=ayeaye.AccessMode.WRITE)

        # two data types that can be added
        p = ayeaye.Pinnate({'common_name': 'Angel fish'})
        c.add(p)

        d = {'common_name': 'Grey reef shark'}
        c.add(d)
        c.close_connection()  # flush to disk

        with open(csv_file, 'r', encoding=c.encoding) as f:
            csv_content = f.read()

        expected_content = ('common_name\n'
                            'Angel fish\n'
                            'Grey reef shark\n'
                            )

        self.assertEqual(expected_content, csv_content)

    def test_tsv_basics(self):
        """
        Tab separated, Iterate all the data items and check each row is being yielded as an
        instance of :class:`ayeaye.Pinnate`
        """
        c = TsvConnector(engine_url="tsv://" + EXAMPLE_TSV_PATH)

        monkey_names = ", ".join([monkey.common_name for monkey in c])
        expected = "Goeldi's marmoset, Common squirrel monkey, Crab-eating macaque"
        assert expected == monkey_names

    def test_multi_connector_append(self):
        """
        Add engine_urls at runtime.
        This is a preemptive measure for when the post build lock file (not yet implemented) will
        store runtime changes to the data connections.
        """
        c = MultiConnector(engine_url=['csv://data_1234.csv', 'csv://data_4567.csv'],
                           access=ayeaye.AccessMode.WRITE,
                           )
        c.engine_url.append('csv://data_8910.csv')

        all_urls = [connector.engine_url for connector in c]
        expected_urls = ['csv://data_1234.csv', 'csv://data_4567.csv', 'csv://data_8910.csv']
        self.assertEqual(expected_urls, all_urls)

        # check that late additions to c.engine_url are visible. 'late' means after c.connect()
        # has been called
        another_file = 'csv://data_1112.csv'
        c.engine_url.append(another_file)
        all_urls = [connector.engine_url for connector in c]
        expected_urls.append(another_file)
        self.assertEqual(expected_urls, all_urls)

    def test_csv_encoding(self):
        """
        Specify character encoding in the URL. This test doesn't ensure data conforms.
        """
        c = CsvConnector(engine_url="csv://" + EXAMPLE_CSV_PATH)
        self.assertEqual("utf-8-sig", c.encoding, "Unexpected default encoding")

        c = CsvConnector(engine_url="csv://" + EXAMPLE_CSV_PATH + ";encoding=latin-1")
        self.assertEqual("latin-1", c.encoding, "Can't override default encoding")

    def test_csv_engine_decode(self):

        c = CsvConnector(engine_url="csv:///data/abc.csv")
        a = c.engine_params
        expected_path = "/data/abc.csv"
        if True or os.path.sep != '/':
            expected_path = expected_path.replace('/', os.path.sep)
        self.assertEqual(expected_path, a.file_path)

        c = CsvConnector("csv:///data/abc.csv;encoding=latin-1;start=3;end=100")
        with self.assertRaises(NotImplementedError):
            c.engine_params

        a = c._engine_params
        expected_path = "/data/abc.csv"
        if True or os.path.sep != '/':
            expected_path = expected_path.replace('/', os.path.sep)
        self.assertEqual(expected_path, a.file_path)
        self.assertEqual("latin-1", a.encoding)
        self.assertEqual(3, a.start)
        self.assertEqual(100, a.end)

    def test_json_basics(self):
        c = JsonConnector(engine_url="json://" + EXAMPLE_JSON_PATH)
        self.assertEqual('London', c.data.name)
        self.assertEqual('light intensity drizzle', c.data.weather.description)

    def test_json_write(self):

        data_dir = tempfile.mkdtemp()
        json_file = os.path.join(data_dir, "chips.json")
        c = JsonConnector(engine_url=f"json://{json_file}", access=ayeaye.AccessMode.WRITE)

        good_examples = [('a string', '"a string"'),
                         (ayeaye.Pinnate({'a': 1}), '{"a": 1}'),
                         ({'a': 1}, '{"a": 1}'),
                         ([1, 2, 3], '[1, 2, 3]')
                         ]

        for acceptable_data, expected_json in good_examples:

            c.data = acceptable_data
            with open(json_file, 'r', encoding=c.encoding) as f:
                json_not_decoded = f.read()
                self.assertEqual(expected_json, json_not_decoded)

        # these data types can't be encoded as JSON
        bad_examples = [set([1, 2, 3]),
                        c,
                        ]
        for unacceptable_data in bad_examples:
            with self.assertRaises(TypeError):
                c.data = unacceptable_data

    def test_csv_missing_values(self):
        """
        Approx position in file not working when None values are in the CSV.
        """
        c = CsvConnector(engine_url="csv://" + EXAMPLE_CSV_BROKEN_PATH)
        current_position = 0
        for _ in c:
            self.assertTrue(c.progress > current_position)
            current_position = c.progress

    def test_csv_without_fieldname_header(self):

        c = CsvConnector(engine_url="csv://" + EXAMPLE_CSV_MICE,
                         field_names=['common_name', 'scientific_name', 'geo_distribution']
                         )
        mice = [mouse.as_dict() for mouse in c]

        expected_line_0 = {'common_name': 'Yellow-necked mouse',
                           'scientific_name': 'Apodemus flavicollis',
                           'geo_distribution': 'Europe'
                           }

        self.assertEqual(3, len(mice))
        # just checking first line is data with correct field names
        self.assertEqual(expected_line_0, mice[0])

    def test_csv_without_fieldname_header_write(self):
        """
        Specify fields. Without this fields are taken from first record to be added.
        """
        data_dir = tempfile.mkdtemp()
        csv_file = os.path.join(data_dir, "lemurs.csv")
        c = CsvConnector(engine_url="csv://" + csv_file, access=ayeaye.AccessMode.WRITE,
                         field_names=['common_name', 'main_colours']
                         )
        for lemur in [{'common_name': 'Indri'},
                      {'common_name': 'Ring tailed', 'main_colours': 'grey, black, white'},
                      ]:
            c.add(lemur)

        c.close_connection()

        with open(csv_file, 'r', encoding=c.encoding) as f:
            csv_content = f.read()

        expected_content = ('common_name,main_colours\n'
                            'Indri,\n'
                            'Ring tailed,"grey, black, white"\n'
                            )

        self.assertEqual(expected_content, csv_content)

    def test_multi_connector_passes_args(self):
        """
        kwargs given to Connect should be passed to each DataConnection created by multi-connect.
        """
        # header-less CSVs
        c = MultiConnector(engine_url=["csv://" + EXAMPLE_CSV_MICE, "csv://" + EXAMPLE_CSV_SQUIRRELS],
                           access=ayeaye.AccessMode.READ,
                           field_names=['common_name', 'scientific_name', 'geo_distribution']
                           )
        rodents = []
        for cx in c:
            rodents.extend([r.as_dict() for r in cx])

        expected_line_0 = {'common_name': 'Yellow-necked mouse',
                           'scientific_name': 'Apodemus flavicollis',
                           'geo_distribution': 'Europe'
                           }

        expected_line_3 = {'common_name': 'American red squirrel',
                           'scientific_name': 'Tamiasciurus hudsonicus',
                           'geo_distribution': 'North America'
                           }

        self.assertEqual(5, len(rodents))
        self.assertEqual(expected_line_0, rodents[0])
        self.assertEqual(expected_line_3, rodents[3])

    def test_json_datasource_exists(self):
        c = JsonConnector(engine_url="json://" + EXAMPLE_JSON_PATH)
        self.assertTrue(c.datasource_exists)

        c = JsonConnector(engine_url="json://this_doesnt_exist.json")
        self.assertFalse(c.datasource_exists)

    def test_multi_connector_by_engine_url(self):
        """
        Get a data connection from a multi connector by engine_url
        """
        engine_0 = "csv://" + EXAMPLE_CSV_MICE
        engine_1 = "csv://" + EXAMPLE_CSV_SQUIRRELS
        c = MultiConnector(engine_url=[engine_0, engine_1],
                           access=ayeaye.AccessMode.READ,
                           field_names=['common_name', 'scientific_name', 'geo_distribution']
                           )

        dataset = c[engine_0]
        # check access to any dataset property
        self.assertTrue(dataset.engine_params.file_path.endswith('tests/data/mice.csv'))
        self.assertEqual(2, len(c))



import unittest

import ayeaye
from ayeaye.connectors.elasticsearch_connector import ElasticsearchConnector


class TestElasticsearchConnector(unittest.TestCase):

    def test_engine_decode(self):

        samples = [('localhost', None, None, "elasticsearch://localhost/"),
                   ('127.0.0.1', None, None, "elasticsearch://127.0.0.1"),
                   ('es_server', 9200, 'my_docs', "elasticsearch://es_server:9200/my_docs"),
                   ]
        for expected_host, expected_port, expected_index, engine_url in samples:
            c = ElasticsearchConnector(engine_url=engine_url)
            host, port, default_index = c._decode_engine_url()
            self.assertEqual(expected_host, host)
            self.assertEqual(expected_port, port)
            self.assertEqual(expected_index, default_index)

    @unittest.skip("Needs integration test suite")
    def test_read_write_document(self):
        "Happy ES paths"

        c = ElasticsearchConnector(engine_url="elasticsearch://localhost/my_docs_xyz",
                                   access=ayeaye.AccessMode.READWRITE
                                   )
        c.add({'hello': 'world'})
        # TODO - search and find a document without an id

        c.add({'hello': 'there'}, doc_id=123)
        result = c.fetch(doc_id=123)
        self.assertEqual({'hello': 'there'}, result, 'Without index or document type should match')

        c.add({'hello': 'Bob'}, doc_id='124', document_type='hr_records', index='my_docs_abc')
        result = c.fetch(doc_id='124', document_type='hr_records', index='my_docs_abc')
        self.assertEqual({'hello': 'Bob'}, result, 'With index and document type should match')

    @unittest.skip("Needs integration test suite")
    def test_missing_index(self):

        c = ElasticsearchConnector(engine_url="elasticsearch://localhost/",
                                   access=ayeaye.AccessMode.WRITE
                                   )
        with self.assertRaises(ValueError) as context:
            c.add({'hello': 'there'}, doc_id=123)
        self.assertIn("Unknown index", str(context.exception))



import os
import unittest

from ayeaye.common_pattern.manifest import EngineFromManifest
from ayeaye.connect_resolve import connector_resolver
from ayeaye.connect import Connect
from ayeaye.model import Model

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_DATA = os.path.join(PROJECT_TEST_PATH, "..", "data")


class TestManifest(unittest.TestCase):

    def test_engine_from_manifest(self):
        """
        Use list of files from manifest to load other datasets.

        @see notes in EngineFromManifest

        ./data/manifest_abcd.json contains a list of files, well it could be a list but is just
        one file - 'blue_ants.csv'.

        'abcd' is the build serial number.
        """
        class InsectSurvey(Model):
            manifest = Connect(engine_url=f"json://{TEST_DATA}/manifest_" + "{build_id}.json")
            ants = Connect(engine_url=EngineFromManifest(manifest, "source_files", "csv"))
            invertebrates = Connect(engine_url=EngineFromManifest(manifest, "single_file", "json"))

            def build(self):
                return

        with connector_resolver.context(build_id="abcd"):
            m = InsectSurvey()
            m.go()  # uses pre_build(), build() etc.
            ants_engine_url = m.ants.engine_url
            invertebrates_engine_url = m.invertebrates.engine_url

        self.assertEqual(ants_engine_url, ["csv://blue_ants.csv"])
        self.assertEqual(invertebrates_engine_url, "json://worms.json")



from io import StringIO
import os
import shutil
import tempfile
import unittest

import ayeaye
from ayeaye.connectors import UncookedConnector

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_FILE = os.path.join(PROJECT_TEST_PATH, 'data', 'quote.txt')


class FakeModel(ayeaye.Model):
    quote = ayeaye.Connect(engine_url=f"file://{EXAMPLE_FILE}")

    def build(self):
        self.log(self.quote.data)


class TestUncookedConnector(unittest.TestCase):

    def setUp(self):
        self._working_directory = None

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_read_access(self):
        """Model uses generic connection (file://) to read contents of a file which is written to
        the log.
        """
        m = FakeModel()
        m.log_to_stdout = False
        external_log = StringIO()
        m.set_logger(external_log)
        m.go()

        external_log.seek(0)
        all_the_logs = external_log.read()
        self.assertIn("Perfection is achieved", all_the_logs)

    def test_on_demand(self):
        "File handle and file path are available as attributes"
        c = UncookedConnector(engine_url=f"file://{EXAMPLE_FILE}")

        self.assertEqual(EXAMPLE_FILE, c.file_path)

        fh = c.file_handle
        self.assertIsNotNone(fh, "Should be a file handle")

        c.close_connection()

        self.assertNotEqual(fh, c.file_handle, "A new file handle should have been opened")

    def test_write(self):

        a_quote = ("More data does not necessarily mean better information."
                   "Bruce Schnier - Cryptogram Oct. 2013"
                   )

        data_dir = tempfile.mkdtemp()
        quotes_file = os.path.join(data_dir, "quotes.notes")

        c = UncookedConnector(engine_url="file://" + quotes_file, access=ayeaye.AccessMode.WRITE)
        c.data = a_quote
        c.close_connection()

        with open(quotes_file, 'r', encoding=c.encoding) as f:
            file_content = f.read()

        self.assertEqual(a_quote, file_content)



'''
Created on 22 Jan 2020

@author: si
'''
try:
    from sqlalchemy import create_engine
    from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
    from sqlalchemy.orm import sessionmaker
    from sqlalchemy.sql import text
except:
    pass

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class SqlAlchemyDatabaseConnector(DataConnector):
    engine_type = ['sqlite://', 'mysql://', 'mysql+pymysql://', 'postgresql://']
    optional_args = {'schema_builder': None,
                     'schema_model': None,
                     }

    def __init__(self, *args, **kwargs):
        """
        Connector to relational databases supported by SQLAlchemy.
        https://www.sqlalchemy.org/

        Supports SQLAlemy's ORM (Object Relational Mapper)

        For args: @see :class:`connectors.base.DataConnector`

        additional args for SqlalchemyDatabaseConnector
            schema_builder (optional) (callable) taking declarative base as the single argument.
            Must return a list of classes or single class with the shared declarative base.
            see https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/api.html

            The behaviour of this connection will differ on single or list being passed.
            For example:
              .schema requires the name of the class in multiple mode.
              e.g.  my_connection.schema.OrmClass (for multiple mode)
                    and my_connection.schema (for 'single schema mode')

            Can't be passed alongside `schema_model`.

            schema_model (optional) SqlAlchemy Model (subclass of
                function:`sqlalchemy.ext.declarative.declarative_base`). Can be passed along-side
                `schema_builder`.

        Connection information-
            engine_url format varied with each engine
        e.g. sqlite:////data/sensors.db  for SQLite DB stored in '/data/sensors.db'
        """
        super().__init__(*args, **kwargs)

        if (self.schema_builder is not None) and (self.schema_model is not None):
            msg = "`schema_builder` and `schema_model` arguments are mutually exclusive"
            raise ValueError(msg)

        # on demand see :method:`connect`
        # the engine, declarative base class and sessions belong to this connection. Sharing these
        # between connections within a single model is not yet implemented.
        self.Base = None  # the declarative base
        self.session = None
        self.engine = None

        # self.schema_builder is built by init from the optional args
        self._schema_p = None  # see :method:`connect`

    def connect(self):

        if self.Base is not None:
            return

        if self.schema_model is not None:

            if isinstance(self.schema_model, list):
                check_schema_models = self.schema_model
            else:
                check_schema_models = [self.schema_model]

            if len(check_schema_models) == 0:
                raise TypeError("No models passed")

            def get_declarative_base(schema_model):
                """Given a class that inherits from SqlAlchemy's declarative base/meta;
                return the class
                """
                if not isinstance(schema_model, DeclarativeMeta):
                    raise TypeError("Not an SqlAlchemy database model")

                # Using method resolution order, find the declarative_base that would be common to
                # all models.
                # e.g. in the following, find Base given MyModel
                #   Base = declarative_base()
                #   ...
                #   class MyModel(Base):
                #       ...
                #
                # in single schema_model mode I'm not sure if common ancestor matters
                resolution_order = schema_model.mro()
                resolution_order.reverse()
                for m in resolution_order:
                    if isinstance(m, DeclarativeMeta):
                        return m
                else:
                    raise ValueError("DeclarativeMeta not found in schema_model??")

            bases = [get_declarative_base(s) for s in check_schema_models]
            for idx in range(len(bases)):
                if idx == 0:
                    continue
                if bases[0] != bases[idx]:
                    msg = "Models passed to `schema_model` must share the same declarative base"
                    raise ValueError(msg)

            self.Base = bases[0]
            self.engine = create_engine(self.engine_url)
            self.Base.metadata.bind = self.engine
            DBSession = sessionmaker(bind=self.engine)
            self.session = DBSession()

            if isinstance(self.schema_model, list):
                as_dict = {c.__name__: c for c in self.schema_model}
                self._schema_p = Pinnate(as_dict)
            else:
                self._schema_p = self.schema_model  # single class

        else:
            # SQL direct or with self.schema_builder callable

            self.Base = declarative_base()
            self.engine = create_engine(self.engine_url)

            # Bind the engine to the metadata of the Base class so that the
            # declaratives can be accessed through a DBSession instance
            self.Base.metadata.bind = self.engine

            DBSession = sessionmaker(bind=self.engine)
            # A DBSession() instance establishes all conversations with the database
            # and represents a "staging zone" for all the objects loaded into the
            # database session object. Any change made against the objects in the
            # session won't be persisted into the database until you call
            # session.commit(). If you're not happy about the changes, you can
            # revert all of them back to the last commit by calling
            # session.rollback()
            self.session = DBSession()

            if self.schema_builder is not None:
                # initialise schema
                schema_classes = self.schema_builder(self.Base) \
                    if self.schema_builder is not None else []

                if isinstance(schema_classes, list):
                    as_dict = {c.__name__: c for c in schema_classes}
                    self._schema_p = Pinnate(as_dict)
                else:
                    self._schema_p = schema_classes  # single class

    def close_connection(self):
        if self.session is not None:
            self.session.close()

    def __del__(self):
        """
        SqlAlchemy does it's own deconstruction.
        """
        pass

    def create_table_schema(self):
        """
        Create the tables defined in self.schema
        """
        if self.access == AccessMode.READ:
            raise ValueError("Can not build schema when access == READ")

        self.connect()
        self.Base.metadata.create_all(self.engine)

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        """
        Generator for all records in all schema.
        """
        if self.access not in [AccessMode.READ, AccessMode.READWRITE]:
            raise ValueError("Can not read data without access == READ")

        schemata = [self.schema] if self.is_single_schema_mode else self.schema.values()
        for schema in schemata:
            # TODO take primary key from schema or default to 'id'
            for r in self.session.query(schema).order_by(schema.id).all():
                yield r

    @property
    def data(self):
        raise NotImplementedError("TODO")

    @property
    def schema(self):
        """
        SQLAlchemy's ORM classes represent tables in the underlying database.

        This property might be pushed up to :class:`DataConnector` so all Connectors could implement schemas.

        :returns: instance of :class:`AyeAye.pinnate` with names of the ORM classes as keys and the
        class (not instance) as the value.
        """
        self.connect()
        return self._schema_p

    @property
    def is_single_schema_mode(self):
        return not isinstance(self.schema, Pinnate)

    def add(self, item):
        """
        @param item: (dict or ORM instance) - dict only with 'single schema mode'
        """
        self.connect()
        if isinstance(item, dict):
            if not self.is_single_schema_mode:
                raise ValueError("Dictionary can only be used in single schema mode")
            item = self.schema(**item)
        else:
            if not isinstance(item, tuple(self.schema.values())):
                msg = "Item of type {} isn't part of this connection's schema"
                raise ValueError(msg.format(type(item)))

        self.session.add(item)

    def commit(self):
        """
        Send pending data changes to the database.
        """
        # only actually needed if no items were added. otherwise the session is already there.
        self.connect()

        # TODO auto commit
        self.session.commit()

    def sql(self, sql_stmt, **sql_params):
        """
        Execute an SQL query against the database engine directly without using the ORM.
        Parameters in the statement are named and the values are passed in a dictionary. Numbered
        positions aren't supported.

        Example- also converts to dictionary
            results = my_db_connector.sql("SELECT * from nice_colours where colour <> :not_really_a_colour",
                                          not_really_a_colour='black'
                                         )
            for r in results:
              print(dict(r))


        @param sql_stmt: (str) SQL statement with
        @param sql_params: (key value pairs), keys must be strings of parameters in SQL statement
        @return: (named tuples) if the query returns results.
        """
        self.connect()
        sql = text(sql_stmt)
        return self.session.execute(sql, sql_params)



import unittest

from ayeaye.connect_resolve import ConnectorResolver, connector_resolver
from ayeaye.connect import Connect
from ayeaye.connectors.csv_connector import CsvConnector
from ayeaye.model import Model


class FakeModel:
    insects = Connect(engine_url="fake://bugsDB")

    def __init__(self):
        self._connections = {}


class TestResolve(unittest.TestCase):
    """
    Runtime variables made available through ayeaye.connect_resolve.connector_resolver.
    """

    def test_multi_connector_resolve(self):
        """
        MultiConnector + ConnectorResolver.
        Other tests for this in :class:`TestConnectors`.
        """

        def simple_resolver(unresolved_engine_url):
            return unresolved_engine_url.format(**{'data_version': '1234'})

        # A MultiConnector
        c = Connect(engine_url=["csv://my_path_x/data_{data_version}.csv",
                                "csv://my_path_y/data_{data_version}.csv"
                                ]
                    )

        with connector_resolver.context(simple_resolver):
            resolved_engine_urls = [data_conn.engine_url for data_conn in c]

        expected_urls = ['csv://my_path_x/data_1234.csv', 'csv://my_path_y/data_1234.csv']
        self.assertEqual(expected_urls, resolved_engine_urls)

    def test_multi_connector_add(self):
        """
        Use MultiConnector's convenience method for adding engine_urls at run time.
        Also ensure the connector resolver is still being used.
        """
        class FishStocksCollator(FakeModel):
            fish = Connect(engine_url=['csv://{file_location}/pond_1.csv',
                                       'csv://{file_location}/pond_2.csv',
                                       ]
                           )

            def build(self):
                # add a new dataset at runtime
                c = self.fish.add_engine_url('csv://{file_location}/pond_3.csv')
                assert isinstance(c, CsvConnector)
                assert c.engine_url == 'csv:///data/pond_3.csv'

        def file_location_resolver(unresolved_engine_url):
            return unresolved_engine_url.format(**{'file_location': '/data'})

        with connector_resolver.context(file_location_resolver):
            m = FishStocksCollator()
            m.build()
            all_urls = [connector.engine_url for connector in m.fish]

        expected_urls = ['csv:///data/pond_1.csv', 'csv:///data/pond_2.csv',
                         'csv:///data/pond_3.csv',
                         ]
        self.assertEqual(expected_urls, all_urls)

    def test_resolve_engine_url(self):
        """
        The engine_url contains a parameter that is replaced on demand.
        """
        msg = "There are existing resolver callables before the test has started"
        self.assertEqual(0, len(connector_resolver.unnamed_callables), msg)

        class MockFakeEngineResolver:
            "Record when it's used and just substitute {data_version} with '1234'"

            def __init__(self):
                self.has_been_called = False

            def __call__(self, unresolved_engine_url):
                self.has_been_called = True
                return unresolved_engine_url.format(**{'data_version': '1234'})

        c = CsvConnector(engine_url="csv://my_path/data_{data_version}.csv")

        m_resolver = MockFakeEngineResolver()
        with connector_resolver.context(m_resolver):
            self.assertFalse(m_resolver.has_been_called, "Should only be called on demand")
            msg = "One resolver exists during the .context"
            self.assertEqual(1, len(connector_resolver.unnamed_callables), msg)

            self.assertEqual('csv://my_path/data_1234.csv', c.engine_url)

            msg = "Should have been called after engine_url is available"
            self.assertTrue(m_resolver.has_been_called, msg)

        msg = "At end of with .context the MockFakeEngineResolver should have been removed"
        self.assertEqual(0, len(connector_resolver.unnamed_callables), msg)

    def test_attribute_access_to_instances(self):

        class SaladResolver:
            def available_today(self):
                return ["csv://cucumbers.csv", "csv://cress.csv"]

        fresh_salad = SaladResolver()
        with connector_resolver.context(salad=fresh_salad):
            todays_engine_urls = connector_resolver.salad.available_today()

        self.assertEqual(["csv://cucumbers.csv", "csv://cress.csv"], todays_engine_urls)
        self.assertNotIn('salad', connector_resolver._attr, "Post context clean up failed")

    def test_deferred_attribute_access(self):
        """
        If a Connect uses a callable to return engine_urls at runtime and this callable uses
        connector_resolver's named attributes there is a catch 22. -- the resolver needs the
        attribute to be set before the model class is imported. Solution is a deferred call that
        is only evaluated by Connect._prepare_connection
        """
        class InsectSurvey(Model):
            ants = Connect(engine_url=connector_resolver.my_ants.all_the_files(ant_types="red"))

            def build(self):
                assert self.ants.engine_url == "csv://red_ants.csv"

        # ------- at this point ------------
        # without the deferred call this would have failed by here because `importing` InsectSurvey
        # would have evaluated 'ants = Connect(...)'

        class MyFileResolver:
            def all_the_files(self, ant_types):
                if ant_types == "red":
                    return "csv://red_ants.csv"
                raise ValueError("This line should be unreachable in this test")

        files_at_runtime = MyFileResolver()
        with connector_resolver.context(my_ants=files_at_runtime):
            m = InsectSurvey()
            m.build()

    def test_without_with_statement(self):
        """
        In unit tests it's helpful to use the same resolver context across a few methods. For
        example in unittest's setUp, tearDown and the test itself.
        """
        class LizardLocator(FakeModel):
            habitats = Connect(engine_url='csv://{file_location}/habitat.csv')

            def get_the_important_engine_url(self):
                return self.habitats.engine_url

        def file_location_resolver(unresolved_engine_url):
            return unresolved_engine_url.format(**{'file_location': '/data'})

        m = LizardLocator()
        with self.assertRaises(ValueError) as exception_context:
            m.get_the_important_engine_url()

        exception_message = str(exception_context.exception)
        msg = "Without a connector_resolver it shouldn't be possible to get the engine_url"
        self.assertIn("Couldn't fully resolve engine URL", exception_message, msg)
        self.assertIn("Missing template variables are: {file_location}", exception_message)

        # using .start() and .finish() instead of a with statement
        local_context = connector_resolver.context(file_location_resolver)
        local_context.start()

        m = LizardLocator()
        self.assertEqual('csv:///data/habitat.csv', m.get_the_important_engine_url())

        msg = "One resolver exists between .start() and .finish()"
        self.assertEqual(1, len(connector_resolver.unnamed_callables), msg)

        # drop the local context
        local_context.finish()

        self.assertEqual(0, len(connector_resolver.unnamed_callables), msg)

    def test_multiple_unnamed_resolvers(self):
        """
        Using ConnectorResolver directly (no as the global/singleton) does chaining unnamed
        resolvers work?
        """
        class IgnoreMissingDict(dict):
            def __missing__(self, key):
                return '{' + key + '}'

        def a2x(e):
            return e.format_map(IgnoreMissingDict(**{'a': 'x'}))

        def b2y(e):
            return e.format_map(IgnoreMissingDict(**{'b': 'y'}))

        def c2z(e):
            return e.format_map(IgnoreMissingDict(**{'c': 'z'}))

        cr = ConnectorResolver()
        cr.add(a2x, b2y, c2z)
        engine_url = cr.resolve("{a}{b}{c}")
        expected_url = "xyz"
        self.assertEqual(expected_url, engine_url)

    def test_named_variables(self):

        with connector_resolver.context(env_secret_password="supersecret"):
            x = Connect(engine_url="mysql://root:{env_secret_password}@localhost/my_database")
            x.connect_standalone()
            self.assertEqual('mysql://root:supersecret@localhost/my_database', x.engine_url)

    def test_deferred_results_not_held(self):
        """
        Regression test for fix. The results of a callable for engine_url were being persisted on
        relayed_kwargs in :method:`Connect._prepare_connection`. DeferredResolution was being used
        in this case and it's a common pattern.
        """
        class AnimalsSurvey(Model):
            rodents = Connect(
                engine_url=connector_resolver.my_survey.sample_data(rodent_type="mice"))

        class ResolverA:
            def sample_data(self, rodent_type):
                if rodent_type == "mice":
                    return "csv://mice_sample_a.csv"
                raise ValueError("This line should be unreachable in this test")

        files_at_runtime = ResolverA()
        with connector_resolver.context(my_survey=files_at_runtime):
            m = AnimalsSurvey()
            first_call_engine_url = m.rodents.engine_url

        class ResolverB:
            def sample_data(self, rodent_type):
                if rodent_type == "mice":
                    return "csv://mice_sample_b.csv"
                raise ValueError("This line should be unreachable in this test")

        files_at_runtime = ResolverB()
        with connector_resolver.context(my_survey=files_at_runtime):
            m = AnimalsSurvey()
            second_call_engine_url = m.rodents.engine_url

        self.assertNotEqual(first_call_engine_url, second_call_engine_url)

    def test_reset_connect_resolve(self):
        """
        :method:`brutal_reset` can be used to stop a failed unit test effecting another unit test.
        """
        connector_resolver.add(x='y')
        with self.assertRaises(ValueError) as exception_context:
            connector_resolver.add(x='z')

        self.assertIn("Attempted to set existing attribute: x", str(exception_context.exception))

        connector_resolver.brutal_reset()
        connector_resolver.add(x='z')

        # don't leave state for other tests
        connector_resolver.brutal_reset()

    @unittest.skip("Callable kwargs not implemented yet")
    def test_callable_mapper_value(self):

        class CheeseSales(Model):
            products = Connect(engine_url="csv://my_path_x/data_{data_version}.csv")

        def simple_resolver(*args):
            return "deep_fried_brie"

        with connector_resolver.context(data_version=simple_resolver):
            m = CheeseSales()
            resolved_engine_url = m.products.engine_url

        self.assertEqual('csv://my_path_x/data_deep_fried_brie.csv', resolved_engine_url)



'''
Created on 22 Jan 2020

@author: si
'''
import os
import shutil
import tempfile
import unittest

from sqlalchemy import inspect, Column, Integer, String
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base

import ayeaye
from ayeaye.connectors.sqlalchemy_database import SqlAlchemyDatabaseConnector


def fruit_schemas(declarative_base):
    class Pear(declarative_base):
        __tablename__ = 'pear'
        id = Column(Integer, primary_key=True)
        variety = Column(String(250), nullable=False)

    class Bananna(declarative_base):
        __tablename__ = 'bananna'
        id = Column(Integer, primary_key=True)
        variety = Column(String(250), nullable=False)

    return [Pear, Bananna]


def people_schema(declarative_base):
    class Person(declarative_base):
        __tablename__ = 'person'
        id = Column(Integer, primary_key=True)
        surname = Column(String(250), nullable=False)
    return Person


class TestSqlAlchemyConnector(unittest.TestCase):

    def setUp(self):
        self._working_directory = None

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_schema_access_multiple(self):
        """
        list of ORM models in schema
        """
        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://",
                                        schema_builder=fruit_schemas,
                                        )
        # method resolution order
        super_classes = [cc.__name__ for cc in c.schema.Pear.mro()]
        expected = ['Pear', 'Base', 'object']  # class above, declarative base, py obj
        self.assertEqual(expected, super_classes)

    def test_schema_access_single(self):
        """
        One ORM model in schema has different way of accessing with .schema
        """
        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://",
                                        schema_builder=people_schema,
                                        )
        # method resolution order
        super_classes = [cc.__name__ for cc in c.schema.mro()]
        expected = ['Person', 'Base', 'object']  # class above, declarative base, py obj
        self.assertEqual(expected, super_classes)

    def test_create_db_schema(self):

        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://",
                                        schema_builder=fruit_schemas,
                                        access=ayeaye.AccessMode.WRITE,
                                        )

        c.connect()

        # check there aren't any tables in the DB here
        with self.assertRaises(OperationalError) as context:
            c.session.query(c.schema.Pear).all()

        self.assertIn("no such table: pear", str(context.exception))

        c.create_table_schema()

        # but there are tables now (but no data in them)
        all_the_pears = c.session.query(c.schema.Pear).all()
        self.assertIsInstance(all_the_pears, list)
        self.assertEqual(0, len(all_the_pears))

    def test_add_orm_data_single(self):

        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://",
                                        schema_builder=people_schema,
                                        access=ayeaye.AccessMode.READWRITE
                                        )
        c.create_table_schema()

        c.add({'surname': 'Cornwallis'})
        c.add({'surname': 'Brunel'})
        c.commit()

        # read them back
        all_the_people = [f"{p.id} {p.surname}" for p in c]
        expected = "1 Cornwallis 2 Brunel"
        self.assertEqual(expected, " ".join(all_the_people))

    def test_add_orm_data_multiple(self):

        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://",
                                        schema_builder=fruit_schemas,
                                        access=ayeaye.AccessMode.READWRITE
                                        )
        c.connect()
        c.create_table_schema()

        with self.assertRaises(ValueError) as context:
            c.add({'variety': 'Cavendish'})

        self.assertIn("Dictionary can only be used in single schema mode", str(context.exception))

        c.add(c.schema.Pear(variety="D'Anjou"))
        c.add(c.schema.Bananna(variety="Cavendish"))
        c.commit()

        # read back mixed types with primary key values belonging to each table (i.e. both are 1)
        mixed_records = [r.__tablename__ + str(r.id) for r in c]
        expected = "pear1 bananna1"
        self.assertEqual(expected, " ".join(mixed_records))

    def test_on_disk(self):
        """
        All the other tests are in memory. Ensure to disk works.

        This test is also being created because windows is refusing to delete open files so
        confirmation that close_connection() is working was experimented with using lsof under
        OSX. But the file handle isn't left open so can't be part of this test.
        """
        db_file = "{}/fruit.db".format(self.working_directory())
        c = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                        schema_builder=fruit_schemas,
                                        access=ayeaye.AccessMode.READWRITE
                                        )
        c.connect()
        c.create_table_schema()

        c.add(c.schema.Pear(variety="Comice"))
        c.commit()

        c.close_connection()

        self.assertTrue(os.access(db_file, os.R_OK))

    def test_double_close_sqlite(self):
        """
        TODO - can't reproduce the "Cannot operate on a closed database." Sqlite error.
        """
        db_file = "{}/fruit.db".format(self.working_directory())
        c = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                        schema_builder=fruit_schemas,
                                        access=ayeaye.AccessMode.READWRITE
                                        )
        c.connect()
        c.create_table_schema()

        c.add(c.schema.Pear(variety="Williams"))
        # c.commit()

        c.close_connection()
        c.__del__()

    def test_two_databases(self):
        """
        The declarative base is created by each SqlAlchemyDatabaseConnector. Ensure models passed
        to each Connector stay in their own engines.
        """
        db_file = "{}/fruit.db".format(self.working_directory())
        fruit = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                            schema_builder=fruit_schemas,
                                            access=ayeaye.AccessMode.READWRITE
                                            )
        fruit.create_table_schema()

        db_file = "{}/people.db".format(self.working_directory())
        people = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                             schema_builder=people_schema,
                                             access=ayeaye.AccessMode.READWRITE
                                             )
        people.create_table_schema()

        # Tables creates in correct DB
        # ----------------------
        inspector = inspect(fruit.engine)
        fruit_tables = {table_name for table_name in inspector.get_table_names()}
        self.assertEqual({'bananna', 'pear'}, fruit_tables)

        inspector = inspect(people.engine)
        people_tables = {table_name for table_name in inspector.get_table_names()}
        self.assertEqual({'person'}, people_tables)

        # Tables can be used in the normal way
        # ----------------------
        fruit.add(fruit.schema.Pear(variety="Comice"))
        fruit.commit()
        fruit.close_connection()

        people.add({'surname': 'Attenborough'})
        people.commit()
        people.close_connection()

    def test_schema_builder_model_exclusive(self):

        PeoplCls = people_schema(declarative_base=declarative_base())

        with self.assertRaises(ValueError):
            SqlAlchemyDatabaseConnector(engine_url="sqlite:////tmp/wontbecreated.db",
                                        schema_builder=people_schema,
                                        schema_model=PeoplCls,
                                        access=ayeaye.AccessMode.READWRITE
                                        )

    def test_schema_model_single(self):
        """
        Instead of passing a callable (i.e. schema_builder argument) pass an SqlAlchemy model
        which already has a declarative base.
        """
        Base = declarative_base()

        class Rodents(Base):
            __tablename__ = 'rodent'
            id = Column(Integer, primary_key=True)
            species = Column(String(250), nullable=False)

        db_file = "{}/rodents.db".format(self.working_directory())
        rodents = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                              schema_model=Rodents,
                                              access=ayeaye.AccessMode.READWRITE
                                              )
        rodents.create_table_schema()
        rodents.add({'species': 'Yellow-necked mouse'})
        rodents.commit()
        rodents.close_connection()

    def test_schema_model_multiple(self):
        """
        see :method:`` but with a list, same idea as how the schema_builder argument can return a
        single schema or list.
        """
        Base = declarative_base()

        class Cats(Base):
            __tablename__ = 'cat'
            id = Column(Integer, primary_key=True)
            name = Column(String(250))

        class Dogs(Base):
            __tablename__ = 'dog'
            id = Column(Integer, primary_key=True)
            name = Column(String(250))

        db_file = "{}/pets.db".format(self.working_directory())
        pets = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                           schema_model=[Cats, Dogs],
                                           access=ayeaye.AccessMode.READWRITE
                                           )
        pets.create_table_schema()

        with self.assertRaises(ValueError) as context:
            pets.add({'name': 'Lady'})
        self.assertIn("Dictionary can only be used in single schema mode", str(context.exception))

        pets.add(pets.schema.Cats(name="Lady"))
        pets.add(pets.schema.Dogs(name="Lady"))

        pets.commit()
        pets.close_connection()

    def test_schema_model_multiple_bases(self):
        """
        Multiple declarative bases on same Connector should fail.
        """
        BaseX = declarative_base()
        BaseY = declarative_base()

        class Cats(BaseX):
            __tablename__ = 'cat'
            id = Column(Integer, primary_key=True)
            name = Column(String(250))

        class Dogs(BaseY):
            __tablename__ = 'dog'
            id = Column(Integer, primary_key=True)
            name = Column(String(250))

        db_file = "{}/pets.db".format(self.working_directory())
        c = SqlAlchemyDatabaseConnector(engine_url=f"sqlite:///{db_file}",
                                        schema_model=[Cats, Dogs],
                                        access=ayeaye.AccessMode.READWRITE
                                        )
        with self.assertRaises(ValueError) as context:
            c.connect()

        self.assertIn("Models passed to `schema_model` must share the same declarative base",
                      str(context.exception)
                      )

    def test_sql_direct(self):
        """
        SQL queries without SqlAlchemy ORM models.
        """
        c = SqlAlchemyDatabaseConnector(engine_url="sqlite://")
        c.sql("CREATE TABLE nice_colours (colour varchar(20))")
        c.sql("INSERT INTO nice_colours values ('blue'), ('green'), ('black')")
        results = c.sql("SELECT colour FROM nice_colours where colour <> :not_really_a_colour",
                        not_really_a_colour='black'
                        )
        final_colours = set()
        for r in results:
            final_colours.add(dict(r)['colour'])

        assert set(['blue', 'green']) == final_colours



'''
Created on 17 Dec 2020

@author: si
'''
import os
import tempfile
import unittest

import ayeaye
from ayeaye.connectors.ndjson_connector import NdjsonConnector

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_NDJSON_UK_PUBS = os.path.join(PROJECT_TEST_PATH, 'data', 'uk_pubs.ndjson')


class TestNdjsonConnector(unittest.TestCase):

    def test_iterate_over_json_lines(self):
        c = NdjsonConnector(engine_url="ndjson://" + EXAMPLE_NDJSON_UK_PUBS)
        uk_pubs_names = [pub.name for pub in c]
        expected = "The New Flying Horse"  # just check one expected value has been found
        self.assertIn(expected, uk_pubs_names)

    def test_ndjson_write(self):
        """
        Write to a file without using a schema.
        """
        data_dir = tempfile.mkdtemp()
        ndjson_file = os.path.join(data_dir, "frog_fish.ndjson")
        c = NdjsonConnector(engine_url="ndjson://" + ndjson_file, access=ayeaye.AccessMode.WRITE)

        for common_name in ['Warty frogfish', 'Hairy Frogfish']:
            p = ayeaye.Pinnate({'common_name': common_name})
            c.add(p)

        c.close_connection()  # flush to disk

        with open(ndjson_file, 'r', encoding=c.encoding) as f:
            file_content = f.read()

        expected_content = ('{"common_name": "Warty frogfish"}\n'
                            '{"common_name": "Hairy Frogfish"}\n'
                            )
        self.assertEqual(expected_content, file_content)



from datetime import datetime
import unittest

from ayeaye.pinnate import Pinnate


class TestPinnate(unittest.TestCase):
    def test_attrib_and_dict(self):
        a = Pinnate({'my_string': 'abcdef'})
        self.assertEqual(a.my_string, 'abcdef')
        self.assertEqual(a['my_string'], 'abcdef')
        self.assertEqual(a.as_dict(), {'my_string': 'abcdef'})

    def test_recurse(self):
        d = {'my_things': [1, 2, {'three': 3}]}
        a = Pinnate(d)
        p = a.my_things
        self.assertTrue(p[0] == 1 and p[1] == 2 and isinstance(p[2], Pinnate))
        self.assertEqual(p[2].three, 3)

    def test_variable_method_name(self):
        """
        check it's possible to give a variable the same name as an existing method.
        """
        d = {'as_dict': 1}
        p = Pinnate(d)
        self.assertEqual("{'as_dict': 1}", str(p.as_dict()))

    def test_as_json(self):
        d = {'number': 1,
             'string': 'hello',
             'date': datetime.strptime("2020-01-15 10:34:12", "%Y-%m-%d %H:%M:%S"),
             'recurse_list': [{'abc': 'def'}],
             'recurse_dict': {'ghi': {'jkl': 'mno'}
                              },
             }
        p = Pinnate(d)
        as_json = str(p.as_json())
        expected = ('{"number": 1, "string": "hello", "date": "2020-01-15 10:34:12", '
                    '"recurse_list": [{"abc": "def"}], "recurse_dict": {"ghi": {"jkl": "mno"}}}'
                    )
        self.assertEqual(expected, as_json)

    def test_recursive_lists(self):
        "bug found with list inside a list"

        d = {'name': 'Stall Street',
             'type': 'Line',
             'payload': [[-2.3597675, 51.3797509], [-2.359739, 51.3797783]],
             'tags': {'name': 'Stall Street',
                      'oneway': 'yes',
                      'highway': 'unclassified',
                      'motor_vehicle:conditional': 'no @ (10:00-18:00)'
                      }
             }
        p = Pinnate(d)
        # accessing the item triggered the bug
        self.assertEqual(-2.3597675, p.payload[0][0])


        # deeper recurse
        d = {'a': [[[{'b':'hello'}]]]}
        p = Pinnate(d)
        self.assertEqual('hello', p.a[0][0][0].b)



'''
Created on 4 Mar 2020

@author: si
'''
'''
Created on 14 Jan 2020

@author: si
'''
try:
    import pyarrow.parquet as pq
except:
    pass

import os

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class ParquetConnector(DataConnector):
    engine_type = 'parquet://'

    def __init__(self, *args, **kwargs):
        """
        Connector to Apache Parquet data files and (TODO) datasets.
        See https://parquet.apache.org/

        Uses Apache arrow.
        See https://arrow.apache.org/docs/index.html

        WARNING this module is more experimental than you can possibly imagine.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for ParquetConnector
         None

        Connection information-
            engine_url format is parquet://<filesystem absolute path>data.parquet
        """
        super().__init__(*args, **kwargs)

        self.table = None
        self.row_count = None

        if self.access != AccessMode.READ:
            raise NotImplementedError('Write access not yet implemented')

    def connect(self):
        if self.table is None:

            engine_params = self.ignition._decode_filesystem_engine_url(self.engine_url)
            file_path = engine_params.file_path
            if os.path.isdir(file_path):
                raise NotImplementedError("Parquet datasets are not yet supported")

            if not os.path.isfile(file_path) or not os.access(file_path, os.R_OK):
                raise ValueError(f"File '{file_path}' not readable")

            self.table = pq.read_table(file_path)
            self.row_count = self.table.num_rows

    def close_connection(self):
        self.table = None
        self.row_count = None

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()

        if self.row_count is None:
            raise ValueError("Number of rows not loaded from source file")

        self.current_row = 0
        for table_batch in self.table.to_batches():
            big_chunk = table_batch.to_pydict()

            # Check assumptions about chunks and relationship with tables....
            #
            # This is testing the assumption that each column has same number of rows.
            # This probably needs to consider schema as fields presumably could be
            # optional.
            chunk_item_lengths = [len(c) for c in big_chunk.values()]
            chunk_length = chunk_item_lengths[0]
            for a_chunk_length in chunk_item_lengths:
                if a_chunk_length != chunk_length:
                    msg = "Unable to process tables with optional fields."
                    raise NotImplementedError(msg)

            # assumption: table rows could be split between batches
            for current_index in range(chunk_length):
                row_as_dict = {k: big_chunk[k][current_index] for k in big_chunk.keys()}
                yield Pinnate(data=row_as_dict)
                self.current_row += 1

    @property
    def data(self):
        raise NotImplementedError("TODO")

    def as_pandas(self):
        """
        @return: (Pandas dataframe)
        """
        self.connect()
        return self.table.to_pandas()

    @property
    def schema(self):
        raise NotImplementedError("TODO")

    @property
    def progress(self):
        if self.access != AccessMode.READ or self.row_count is None or self.current_row == 0:
            return None

        return self.current_row / self.file_size



from datetime import datetime
from enum import Enum
from time import time
import warnings

import ayeaye
from ayeaye.connectors.base import DataConnector
from ayeaye.connect_resolve import connector_resolver
from ayeaye.ignition import EngineUrlCase, EngineUrlStatus


class LockingMode(Enum):
    """
    How to capture context around datasets in a model. This is used to track data provenance and
    to make model repeatability possible.

    @see :method:`Model.lock`
    """
    CONTEXT = 'context'  # just the ayeaye.connector_resolver context
    ALL_DATASETS = 'all_datasets'  # the engine_urls for all datasets


class Model:
    """
    Do the thing!

    Abstract class 

    The thing is probably an ETL (Extract, Transform and Load) task which is at the minimum
    the :method:`build`. It could optionally also have a :method:`pre_build_check`, which is run
    first and must succeed. After the :method:`build` an optional :method:`post_build_check`
    can be used to see if build worked. 
    """

    def __init__(self):
        self._connections = {}

        self.log_to_stdout = True
        self.external_logger = None
        self.progress_log_interval = 20  # minimum seconds between progress messages to log

        # stats
        self.start_build_time = None
        self.progress_logged = None  # time last logged

    def go(self):
        """
        Run the model.

        The steps to run a model are
        1. :method:`pre_build_check` - Optional conditional check if model's pre-conditions have
           been met.
        2. :method:`build` - The main modelling stage
        3. :method:`post_build_check` - Optional check if the outputs are valid. e.g. a data
           validation check that is simple and concise. It's like a unit test but on changable
           data.

        Datasets are closed after each stage. This ensures the connections don't have a state, for
        example a position within a file.

        @return: boolean for success
        """
        self.start_build_time = time()
        if not self.pre_build_check():
            self.log("Pre-build check failed", "ERROR")
            self.close_datasets()
            return False
        self.close_datasets()

        self.build()
        self.close_datasets()

        if not self.post_build_check():
            self.log("Post-build check failed", "ERROR")
            self.close_datasets()
            return False
        self.close_datasets()

        return True

    def pre_build_check(self):
        """
        Optionally implemented by subclasses to check any conditions that must be met before
        running :method:`build`.

        For example, assumptions :method:`build` makes on on the format or values within the data
        could be checked here in order to keep code in build simple.

        @return: boolean. If False is returned :method:`build` won't be run.
        """
        return True

    def build(self):
        raise NotImplementedError("All models must implement this method")

    def post_build_check(self):
        """
        This is an optional method that will be run after :method:`build`. It can be used to check
        the validity of the build process.

        @return: boolean. If False is returned the model is considered to have failed.
        """
        return True

    @classmethod
    def connects(cls):
        """
        :returns (dict) of :class:`Connect` classes declared as class variables for this model.
                key is class variable name
                value is :class:`ayeaye.Connect`
        """
        # find :class:`ayeaye.Connect` connections to datasets
        connects = {}
        for obj_name in dir(cls):
            obj = getattr(cls, obj_name)
            if isinstance(obj, ayeaye.Connect):
                connects[obj_name] = obj

        return connects

    def datasets(self):
        """
        :returns (dict) of dataset connections for this model.
                key is class variable name
                value is :class:`ayeaye.DataConnector`
        """
        # find :class:`ayeaye.Connect` connections to datasets
        connections = {}
        for obj_name in dir(self):
            obj = getattr(self, obj_name)
            if isinstance(obj, DataConnector):
                connections[obj_name] = obj

        return connections

    def close_datasets(self):
        """
        Call :method:`close_connection` on all datasets.
        """
        for connection in self.datasets().values():
            connection.close_connection()

    def set_logger(self, logger):
        """
        Also log to something with a :method:`write`.
        e.g. StringIO
        """
        self.external_logger = logger

    def log(self, msg, level="INFO"):
        """
        @param level: (str) DEBUG, PROGRESS, INFO, WARNING, ERROR or CRITICAL
        """
        if not (self.log_to_stdout or self.external_logger is not None):
            return

        date_str = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        msg = "{} {}{}".format(date_str, level.ljust(10), msg)

        if self.external_logger is not None:
            self.external_logger.write(msg)

        if self.log_to_stdout:
            print(msg)

    def log_progress(self, position_pc, msg=None):
        """
        Externally provided indication of position through build is used to calculate ETA (TODO)
        and occasionally log progress to show user something is happening.

        @param position_pc: (float) or None meaning ignore
        @param msg: (str) additional user friendly info
        """
        time_now = time()

        if position_pc > 0.0001 and \
            (self.progress_logged is None or
             self.progress_logged + self.progress_log_interval < time_now):

            if msg is None:
                msg = ""

            progress_pc = "{:.2f}%".format(position_pc * 100)
            running_secs = time_now - self.start_build_time
            eta = "{:.2f}".format((running_secs / position_pc) - running_secs)
            self.log(f"{progress_pc} {eta} seconds remaining. {msg}", level="PROGRESS")
            self.progress_logged = time_now

    def fetch_locking(self):
        """
        A 'lock' is the information needed by a model to reproduce an output.

        This method is optionally implemented by subclasses that need more than the context
        provided by `ayeaye.connector_resolver` in order to be repeatable. For example, the model
        might not be deterministic.

        The value returned will be passed to :method:`apply_locking` if/when a model needs to
        be repeated.

        @returns something that can be serialised to JSON
        """
        return None

    def apply_locking(self, lock_doc):
        """
        Use the output from :method:`fetch_locking` to reproduce results from a previous build.

        This method should be implemented if a model needs to be hydrated with locking details from
        a previous build.
        """
        raise NotImplementedError("Missing sub-class method on attempt to re-hydrate locking.")

    def lock(self, locking_level=LockingMode.CONTEXT):
        """
        A 'lock' is the information needed by a model to reproduce an output.

        Return a JSON safe dictionary of variables needed to repeat the build.

        The returned dictionary is expected to be serialised and stored.

        @param locking_level: (LockingMode)        
            LockingMode.CONTEXT - key values from ayeaye.connector_resolver
            LockingMode.ALL_DATASETS - CONTEXT + engine_urls from all datasets in model.

        @return (dict)
        """
        # this is very much work in progress.
        # coming soon
        # - code version (git commitishes)
        # - library code versions (maybe pipenv lock file)

        locking_doc = {'resolve_context': connector_resolver.capture_context()
                       }

        model_lock = self.fetch_locking()
        if model_lock is not None:
            locking_doc['model_locking'] = model_lock

        if locking_level == LockingMode.ALL_DATASETS:
            locking_doc['dataset_engine_urls'] = {}
            for dataset_name, connector in self.datasets().items():

                # TODO: this should be EngineUrlCase.WITHOUT_SECRETS. Secrets shouldn't be in
                # locking doc. But that's not implemented yet,
                status, engine_url = connector.ignition.engine_url_at_state(
                    EngineUrlCase.FULLY_RESOLVED)
                msg = ("Incomplete implementation: if there are secrets in the engine_url they "
                       "will be included in locking doc."
                       )
                warnings.warn(msg)

                if status != EngineUrlStatus.OK:
                    raise ValueError(f"Can't lock, engine_url not available for '{dataset_name}'")

                locking_doc['dataset_engine_urls'][dataset_name] = engine_url

        return locking_doc



from datetime import datetime
import unittest

from ayeaye.connectors.kafka_connector import KafkaConnector

EXAMPLE_ENGINE_URL_0 = "kafka://localhost/topic=foobar;start=@(2019-05-15 08:00:00);end=@(2019-05-15 18:00:00);"
EXAMPLE_ENGINE_URL_1 = "kafka://localhost/topic=foobar;start=@(2019-05-15 15:50:18);end=@(2019-05-15 15:50:24);"
EXAMPLE_ENGINE_URL_2 = "kafka://localhost/topic=uas;start=@(2019-05-22 10:42:00);end=@(2019-05-22 10:44:00);"


class TestKafkaConnector(unittest.TestCase):

    def test_engine_decode(self):
        date_format = "%Y-%m-%d %H:%M:%S"
        c = KafkaConnector(engine_url=EXAMPLE_ENGINE_URL_0)
        bootstrap_server, topic, start_params, end_params = c._decode_engine_url()
        assert bootstrap_server == 'localhost'
        assert topic == 'foobar'
        assert start_params == datetime.strptime("2019-05-15 08:00:00", date_format)
        assert end_params == datetime.strptime("2019-05-15 18:00:00", date_format)

    @unittest.skip("Needs integration test suite")
    def test_partition_ranges(self):
        c = KafkaConnector(engine_url=EXAMPLE_ENGINE_URL_1)
        p_ranges = set([x for x in c._partition_ranges()])
        # my values right now, need to be aligned with test data when there's an integration env
        expected = {(0, 0, 233303), (1, 0, 133271), (2, 0, 133526)}
        assert expected == p_ranges

    @unittest.skip("Needs integration test suite")
    def test_subset_of_items(self):
        """
        Use a date range to get some items from a topic.
        """
        c = KafkaConnector(engine_url=EXAMPLE_ENGINE_URL_2)
        some_items = [x for x in c]
        # TODO there is a problem with .offsets_for_times(tx)
        # there should be a few hundred items here, not thousands
        assert len(some_items) == 100

    @unittest.skip("Needs integration test suite")
    def test_partition_ranges_2(self):
        c = KafkaConnector(engine_url=EXAMPLE_ENGINE_URL_2)
        p_ranges = set([x for x in c._partition_ranges()])
        # my values right now, need to be aligned with test data when there's an integration env
        expected = {(0, 0, 91)}
        assert expected == p_ranges



'''
Created on 16 May 2019

@author: parkes25
'''
from datetime import datetime
from typing import Generator

try:
    from kafka import KafkaConsumer, KafkaProducer, TopicPartition
    from kafka.structs import OffsetAndTimestamp
except:
    pass

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class KafkaConnector(DataConnector):
    engine_type = 'kafka://'

    def __init__(self, *args, **kwargs):
        """
        Connector to Apache Kafka.
        Args: @see :class:`connectors.base.DataConnector`

        additional args for KafkaConnector
         None

        Connection information-
            engine_url format is kafka://bootstrap_server/topic=<topic>;[start params;][end params;]
        start and end params can be partitions with offsets or '@' notation to use dates.
        e.g. kafka://bionic/topic=foobar;start=@(2019-05-15 08:00:00);end=@(2019-05-15 18:00:00);
        """
        super().__init__(*args, **kwargs)

        # set by :method:`connect`
        self.bootstrap_server = self.topic = self.start_params = self.end_params = None
        self.start_p_offsets = self.end_p_offsets = None
        self.available_topics = None
        self.client = None

        # publicly readable
        self.stats = Pinnate({'added': 0})

        # used during read
        self.approx_position = None
        self.items_to_fetch = None

    def close_connection(self):
        if self.access == AccessMode.WRITE and self.client is not None:
            self.flush()

    def connect(self):
        if self.client is None:
            self.bootstrap_server, self.topic, self.start_params, self.end_params = \
                self._decode_engine_url()

            if self.access == AccessMode.READ:
                self.client = KafkaConsumer(bootstrap_servers=self.bootstrap_server)
                self._setup_consumer()

            elif self.access == AccessMode.WRITE:
                if self.start_params is not None or self.end_params is not None:
                    raise ValueError("Start and end offsets can't be set when writing")
                self.client = KafkaProducer(bootstrap_servers=self.bootstrap_server)

            else:
                raise NotImplementedError('Unknown access mode')

    def _setup_consumer(self):
        """
        prepare offset numbers etc. for reading from Topic
        """
        # <WTF> https://github.com/dpkp/kafka-python/issues/601
        self.available_topics = self.client.topics()
        # </WTF>

        # might as well use it
        assert self.topic in self.available_topics

        if (self.start_params is None) != (self.end_params is None):
            raise ValueError("Both start and end params must be set or both must be None")

        if self.start_params is None:
            # setup partitions to read through
            # TODO not checked with multiple partitions since inheriting from foxglove
            # An offset is assigned to make repeatability (via a locking file) possible later on.
            # and it's easier to terminate the fetch loop this way.
            p_id = self.client.partitions_for_topic(self.topic)
            topic_partitions = [TopicPartition(topic=self.topic, partition=p) for p in list(p_id)]
            starts = self.client.beginning_offsets(topic_partitions)
            ends = self.client.end_offsets(topic_partitions)

            self.start_p_offsets = {tp: OffsetAndTimestamp(
                offset=offset, timestamp=None) for tp, offset in starts.items()}
            self.end_p_offsets = {tp: OffsetAndTimestamp(
                offset=offset - 1, timestamp=None) for tp, offset in ends.items()}

        else:
            # TODO - this code was inherited from Foxglove and hasn't be checked through
            # setup start and end partitions and offsets
            # self.client.seek_to_beginning()
            # datetime is only start/end implemented
            assert isinstance(self.start_params, datetime) and isinstance(self.end_params, datetime)
            start = int(self.start_params.timestamp() * 1000)
            end = int(self.end_params.timestamp() * 1000)

            partitions = self.client.partitions_for_topic(self.topic)
            tx = {TopicPartition(topic=self.topic, partition=p): start
                  for p in list(partitions)}
            self.start_p_offsets = self.client.offsets_for_times(tx)

            # if you give a timestamp after the last record it returns None
            for tp, offset_details in self.start_p_offsets.items():
                if offset_details is None:
                    raise ValueError("Start date outside of available messages")

            tx = {TopicPartition(topic=self.topic, partition=p): end
                  for p in list(partitions)}
            self.end_p_offsets = self.client.offsets_for_times(tx)

            # as above - out of range, for end offset give something useful
            for tp, offset_details in self.end_p_offsets.items():
                if offset_details is None:
                    # go to last message. I'm not 100% sure this is correct
                    end_offsets = self.client.end_offsets([tp])
                    offset = end_offsets[tp] - 1
                    self.end_p_offsets[tp] = OffsetAndTimestamp(offset=offset, timestamp=None)

    def _decode_engine_url(self):
        """
        Returns:
            bootstrap_server, topic, start_params, end_params
            bootstrap_server and topic are (str)
            start_params, end_params are (None), (datetime) or (mixed) - not implemented
                                but will be partition+offsets pairs
        """
        date_format = "%Y-%m-%d %H:%M:%S"
        r = dict(topic=None, start=None, end=None)
        s_url = self.engine_url[len(self.__class__.engine_type):]
        bootstrap_server, r_url = s_url.split('/', 1)
        for param_section in r_url.split(';'):
            if len(param_section) == 0:
                continue
            k, v = param_section.split('=', 1)
            if k in r:
                r[k] = v
        # resolve to dates if needed
        # partition+offset not implemented so start and end must be None or start with @
        # to resolve to a datetime.
        for position in ('start', 'end'):
            p_marker = r[position]
            if p_marker is not None:
                assert p_marker.startswith('@(') and p_marker.endswith(')')
                date_str = p_marker[2:-1]
                r[position] = datetime.strptime(date_str, date_format)

        return bootstrap_server, r['topic'], r['start'], r['end']

    def schema(self):
        raise NotImplementedError("TODO")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def _partition_ranges(self) -> Generator:
        """
        yield partition (int), start_offset (int), end_offset (int)
        for range given in self.engine_url
        """
        self.connect()
        for topic_partition, start_offset_time in self.start_p_offsets.items():
            end_offset_time = self.end_p_offsets[topic_partition]
            yield topic_partition.partition, start_offset_time.offset, end_offset_time.offset

    @property
    def data(self) -> Generator:
        """
        Generator yielding just the value of the record from Kafka.
        Value is made into Pinnate object.

        See https://kafka-python.readthedocs.io/en/master/apidoc/KafkaConsumer.html
        useful attribs include
        m.offset, m.partition, m.timestamp, m.key, m.value
        """
        # Not using a consumer group and setting partitions manually so it's a smaller
        # jump to make this deterministic/repeatable with multiple workers later on.

        self.connect()

        self.approx_position = 0
        for partition_id, start_offset, end_offset in self._partition_ranges():

            # TODO - confirm this can never jump to another partition
            tp = TopicPartition(topic=self.topic, partition=partition_id)
            self.client.assign([tp])

            self.items_to_fetch = end_offset - start_offset
            self.client.seek(tp, start_offset)

            if self.items_to_fetch <= 0:
                msg = f"Invalid offsets {start_offset}:{end_offset} for partition {partition_id}"
                raise ValueError(msg)

            for m in self.client:

                self.approx_position += 1
                yield Pinnate(data=m.value)

                if end_offset is not None and m.offset >= end_offset:
                    break

    def add(self, data, partition=None):
        """
        Write message to topic.
        @param data: (str)
        @param partition: (int) Kafka partition. Not yet implemented.
        """
        # TODO expand data to include binary and instance of :class:`Pinnate` but needs a way of
        # de-serialising on retrieve.

        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        if partition is not None:
            raise NotImplementedError("Placeholder value, not implemented yet")

        if not isinstance(data, str):
            raise ValueError("data isn't an accepted type. Only (str) is accepted.")

        self.connect()

        # TODO use futures
        self.client.send(self.topic, value=bytes(data, 'utf-8'))
        self.stats.added += 1

    def flush(self):
        """
        Ensure all messages have been sent to Kafka
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Flush attempted on dataset opened in READ mode.")

        # TODO futures and performance stats
        self.client.flush()

    @property
    def progress(self):
        if self.access != AccessMode.READ \
                or self.items_to_fetch is None \
                or self.approx_position is None:
            return None

        return self.approx_position / self.items_to_fetch



'''
Created on 19 May 2021

@author: si
'''
import os

from ayeaye.connectors.base import DataConnector, AccessMode


class UncookedConnector(DataConnector):
    engine_type = 'file://'

    def __init__(self, *args, **kwargs):
        """
        Connector to use a raw local file.

        The file will be opened on demand. i.e. reading .data or getting the .file_handle. Getting
        the .file_path will not open the file.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for UncookedConnector
         None

        TODO binary open

        Connection information-
            engine_url format is
            file://<filesystem absolute path>[;encoding=<character encoding>]
        e.g. file:///data/my_project/interesting_notes.txt;encoding=latin-1
        """
        super().__init__(*args, **kwargs)

        self._reset()

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError('READWRITE access not yet implemented')

    def _reset(self):
        self._file_handle = None  # lazy eval, use self.file_handle
        self._encoding = None
        self._engine_params = None
        self.file_size = None

    @property
    def file_handle(self):
        """
        File handle to open file for operations such as :method:`read`()
        """
        self.connect()
        return self._file_handle

    @property
    def file_path(self):
        """
        @return: (str) filesystem path to file
        """
        return self.engine_params.file_path

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding
        """
        if self._engine_params is None:
            self._engine_params = self.ignition._decode_filesystem_engine_url(
                self.engine_url,
                optional_args=['encoding']
            )

            if 'encoding' in self._engine_params:
                self._encoding = self.engine_params.encoding

        return self._engine_params

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if 'encoding' in ep else 'utf-8-sig'

        return self._encoding

    def close_connection(self):
        if self._file_handle is not None:
            self._file_handle.close()
        self._reset()

    def connect(self):
        if self._file_handle is None:

            if self.access == AccessMode.READ:
                self._file_handle = open(self.engine_params.file_path, 'r', encoding=self.encoding)
                self.file_size = os.stat(self.engine_params.file_path).st_size

            elif self.access == AccessMode.WRITE:
                self._file_handle = open(self.engine_params.file_path, 'w', encoding=self.encoding)

            else:
                raise ValueError('Unknown access mode')

    def __len__(self):
        """
        @return: (int) file size
        """
        if self.access != AccessMode.READ:
            raise NotImplementedError("Not yet available in write mode")

        self.connect()
        return self.file_size

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        "Iteration not possible"

        # MAYBE make a readlines method?
        raise TypeError("Can't iterate through a file")

    @property
    def data(self):
        """
        @return: contents of entire file. Type depends on encoding arguments.
        """
        if self.access != AccessMode.READ:
            raise ValueError("Not open in read mode")

        file_content = self.file_handle.read()
        return file_content

    @data.setter
    def data(self, file_content):
        """
        attribute based setter.

        e.g.
        c = UncookedConnector(engine_url="file://mydata_file", access=ayeaye.AccessMode.WRITE)
        c.data = "hello world!"
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Not open in write mode")

        file_content = self.file_handle.write(file_content)

    @property
    def schema(self):
        raise TypeError("Uncooked connectors don't have rich interfaces like schemas")



import unittest

from ayeaye import Connect, connector_registry
from ayeaye.connectors.base import DataConnector

class FakeConnector(DataConnector):
    engine_type = 'superdata://'

    @property
    def data(self):
        return "fake data"

class TestConnectorsRegistry(unittest.TestCase):

    def tearDown(self):
        # ensure private connectors aren't left available outside of each test
        connector_registry.reset()
    
    def test_unknown_engine(self):
        "Connector not builtin to ayeaye and not registered"
        
        c = Connect(engine_url="magicdata://ref_to_data")
        with self.assertRaises(NotImplementedError) as context:
            c.non_existant_attrib_loaded_on_demand
        
        self.assertEqual("Unknown engine in url:magicdata://ref_to_data", str(context.exception))

    def test_non_public_engine(self):
        
        connector_registry.register_connector(FakeConnector)
        
        super_data = Connect(engine_url="superdata://ref_to_data")
        self.assertEqual("fake data", super_data.data)

    def test_valid_registrations(self):
        
        # should be a class
        fake_connector_obj = FakeConnector()
        with self.assertRaises(TypeError):
            connector_registry.register_connector(fake_connector_obj)
        
        # should be a subclass of DataConnector
        class X:
            pass

        with self.assertRaises(TypeError):
            connector_registry.register_connector(X)



from tempfile import TemporaryFile
from typing import Generator, Tuple

try:
    from google.auth.credentials import Credentials
    from google.cloud import storage
    from google.cloud.storage import Blob
except:
    pass

from ayeaye.connectors.base import DataConnector
from ayeaye.connectors.flowerpot import FlowerpotEngine


class GcsFlowerpotConnector(DataConnector):
    engine_type = 'gs+flowerpot://'
    optional_args = {'credentials': None}

    def __init__(self, *args, **kwargs):
        """
        Connector to Datalab :class:`Flowerpot` stored in Google Cloud Storage (GCS) buckets.
        Args: @see :class:`connectors.base.DataConnector`

        additional args for GcsFlowerpotConnector
         'credentials' : (dict) for access within Google Cloud Platform

            engine_url format is gs+flowerpot://[project.]<bucket>/<path>
                    e.g. gs+flowerpot://my_project.my_bucket/some/data/file.flowerpot
                    where gs://my_bucket/some/data/file.flowerpot is the gsutil style URL
        """
        super().__init__(*args, **kwargs)

        self.project_name, self.bucket_name, self.flowerpot_path = \
            split_gcs_uri(kwargs['engine_url'])
        self.bucket = get_gcs_bucket(self.bucket_name, self.project_name, self.credentials)
        self._flowerpot = None # loaded on demand

    def _get_blob(self):
        """Get flowerpot object from its remote path as a GCS blob"""
        return self.bucket.get_blob(self.flowerpot_path)

    @property
    def flowerpot(self) -> 'LazyFlowerpotReader':
        if self._flowerpot is None:
            flowerpot = self._download_flowerpot()
            self._flowerpot = FlowerpotEngine.from_file(flowerpot)
        return self._flowerpot

    def _download_flowerpot(self) -> TemporaryFile:
        flowerpot_file = TemporaryFile('w+b')
        self._get_blob().download_to_file(flowerpot_file)
        flowerpot_file.seek(0)
        return flowerpot_file

    @property
    def data(self) -> Generator[object, None, None]:
        return self.flowerpot.items()

    @property
    def schema(self):
        return None

    def connect(self):
        # looking at this property forces lazy download
        assert self.flowerpot

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")


def split_gcs_uri(gcs_uri: str) -> Tuple[str, str]:
    """
    Split a Google Cloud Storage URL into component (bucket and key) parts for use with GCS API.
    Args:
        gcs_uri: The `gs://` URI to split
    Returns:
        tuple: (bucket, key) pair of GCS bucket name and GCS key of item in URL.
    """
    url_prefix = GcsFlowerpotConnector.engine_type
    assert gcs_uri.startswith(url_prefix), "Google Cloud Storage URIs should start with {}".format(url_prefix)
    proj_bucket, path = tuple(gcs_uri[len(url_prefix):].split('/', 1))
    if '.' in proj_bucket:
        project, bucket = proj_bucket.split('.', 1)
        return project, bucket, path
    return None, proj_bucket, path


def get_gcs_bucket(bucket_name, project=None, credentials=None):
    """
    Convenience method for getting a GCS Bucket() object from a bucket name
    :param: bucket_name (str)
    :param: project (str)
    :param: credentials (google.auth.credentials,Credentials)
    """
    storage_client = storage.Client(project, credentials)
    return storage_client.get_bucket(bucket_name)



'''
Created on 4 Mar 2020

@author: si
'''
import os
import unittest

import pandas as pd

from ayeaye.connectors.parquet_connector import ParquetConnector

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_HELLO = os.path.join(PROJECT_TEST_PATH, 'data', 'hello.parquet')

"""
Test data built with this-

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

df = pd.DataFrame({'name': ['Alice', 'Bob'],
                   'favorite_colour': ['blue', 'green'],
                  },
                  index=list('ab'))

pq.write_table(table, 'hello.parquet')
"""


class TestSqlAlchemyConnector(unittest.TestCase):

    def test_read_as_rows(self):
        """
        Iterate through a parqet file row by row for all columns.
        """
        c = ParquetConnector(engine_url="parquet://" + EXAMPLE_HELLO)
        all_records = [r.as_dict() for r in c]

        self.assertEqual(2, len(all_records), "There are two sample records")

        # there are a couple of other keys as a result of the pandas index. Just check the
        # payload fields
        wanted = [{'name': 'Alice', 'favorite_colour': 'blue'},
                  {'name': 'Bob', 'favorite_colour': 'green'}
                  ]
        for idx, expected_row in enumerate(wanted):
            for expected_key, expected_value in expected_row.items():
                self.assertEqual(expected_value, all_records[idx][expected_key])

    def test_read_as_pandas(self):
        c = ParquetConnector(engine_url="parquet://" + EXAMPLE_HELLO)
        p = c.as_pandas()

        self.assertIsInstance(p, pd.DataFrame)
        self.assertEqual('Alice', p['name'][0], "Can't find expected value in Pandas dataframe")



import unittest

import ayeaye
from ayeaye.connectors.models_connector import ModelsConnector


class One(ayeaye.Model):
    a = ayeaye.Connect(engine_url="csv://a")
    b = ayeaye.Connect(engine_url="csv://b", access=ayeaye.AccessMode.WRITE)


class Two(ayeaye.Model):
    b = One.b.clone(access=ayeaye.AccessMode.READ)
    c = ayeaye.Connect(engine_url="csv://c", access=ayeaye.AccessMode.WRITE)


class Three(ayeaye.Model):
    c = Two.c.clone(access=ayeaye.AccessMode.READ)
    d = ayeaye.Connect(engine_url="csv://d", access=ayeaye.AccessMode.WRITE)


class Four(ayeaye.Model):
    b_copy_paste = ayeaye.Connect(engine_url="csv://b", access=ayeaye.AccessMode.READ)
    e = ayeaye.Connect(engine_url="csv://e", access=ayeaye.AccessMode.WRITE)


class Five(ayeaye.Model):
    b = One.b.clone(access=ayeaye.AccessMode.READ)
    f = ayeaye.Connect(engine_url="sqlite:////data/f.db", access=ayeaye.AccessMode.READWRITE)


class Six(ayeaye.Model):
    b = One.b.clone(access=ayeaye.AccessMode.READ)
    f = Five.f.clone(access=ayeaye.AccessMode.WRITE)


failed_callable_msg = "No test should be calling this as the parent model class isn't instantiated"


def find_destination():
    """Will be called at build() time. In real life this would find out something that should
    only be looked up during runtime."""
    # return "csv://g.csv"
    raise Exception(failed_callable_msg)


def another_find_destination():
    raise Exception(failed_callable_msg)


class Seven(ayeaye.Model):
    b = One.b.clone(access=ayeaye.AccessMode.READ)
    g = ayeaye.Connect(engine_url=find_destination, access=ayeaye.AccessMode.WRITE)


class Eight(ayeaye.Model):
    g = Seven.g.clone(access=ayeaye.AccessMode.READ)
    h = ayeaye.Connect(engine_url="csv://h", access=ayeaye.AccessMode.WRITE)


class Nine(ayeaye.Model):
    i = ayeaye.Connect(engine_url=another_find_destination, access=ayeaye.AccessMode.WRITE)
    h = ayeaye.Connect(engine_url="csv://h", access=ayeaye.AccessMode.WRITE)


class TestModelConnectors(unittest.TestCase):

    @staticmethod
    def repr_run_order(run_order):
        """
        @return: (list of sets) showing a simplified representation of the run_order from
        :method:`_resolve_run_order` using just the 'model_name' field.
        """
        r = []
        for task_group in run_order:
            assert isinstance(task_group, set)
            name_set = set([t.model_name for t in task_group])
            r.append(name_set)
        return r

    def test_single_standalone_model(self):
        c = ayeaye.Connect(models=One)
        msg = ("Attribute access should be proxied through Connect to an instance of "
               "ModelsConnector which should refer back to the Connect instance that created it."
               )
        self.assertEqual(c, c.connect_instance, msg=msg)
        self.assertEqual(c.models, [One], "Single model should be proxied through Connect.")

    def test_construction(self):
        """
        valid ways to make a ModelsConnector instance.
        """
        m = ModelsConnector(models=One)
        self.assertIsInstance(m.models, list, "Single model becomes run-list")

        m = ModelsConnector(models=[One, Two, Three])
        self.assertIsInstance(m.models, list, "Preserves list")

        m = ModelsConnector(models=set([One, Two, Three]))
        self.assertIsInstance(m.models, set, "Preserves set")

        with self.assertRaises(ValueError):
            # non-model
            m = ModelsConnector(models=[One, Two, Three()])

        def models_choosen_at_runtime():
            return set([One, Two])

        with self.assertRaises(NotImplementedError):
            # TODO
            m = ModelsConnector(models=models_choosen_at_runtime)

    def test_resolve_run_order_linear(self):
        """
        Dataset dependencies used to determine model run order.
        """
        c = ayeaye.Connect(models={One, Two, Three})
        r = c._resolve_run_order()

        leaf_sources = set([c.relayed_kwargs['engine_url'] for c in r.leaf_sources])
        expected_leaf_sources = {"csv://a"}
        self.assertEqual(expected_leaf_sources, leaf_sources)

        leaf_targets = set([c.relayed_kwargs['engine_url'] for c in r.leaf_targets])
        expected_leaf_targets = {"csv://d"}
        self.assertEqual(expected_leaf_targets, leaf_targets)

        msg = "Should be a single linear execution"
        self.assertIsInstance(r.run_order, list, msg)

        self.assertEqual([{'One'}, {'Two'}, {'Three'}], self.repr_run_order(r.run_order), msg)

    def test_resolve_run_order_one_branch(self):
        c = ayeaye.Connect(models={One, Two, Four})
        r = c._resolve_run_order()
        self.assertEqual([{'One'}, {'Two', 'Four'}], self.repr_run_order(r.run_order))

    def test_resolve_run_order_readwrite(self):
        c = ayeaye.Connect(models={One, Two, Five, Six})
        r = c._resolve_run_order()
        msg = ("There is an ambiguity because Six is WRITE and Five is READWRITE to the same "
               "dataset (f). The write only is happening first. Feels correct but might need "
               "more thought."
               )
        self.assertEqual([{'One'}, {'Two', 'Six'}, {'Five'}], self.repr_run_order(r.run_order), msg)

    def test_resolve_with_callable(self):
        "Seven has a callable to build it's engine_url at build time"
        c = ayeaye.Connect(models={One, Eight, Seven})
        r = c._resolve_run_order()
        self.assertEqual([{'One'}, {'Seven'}, {'Eight'}], self.repr_run_order(r.run_order))

    def test_resolve_with_two_different_callables(self):
        c = ayeaye.Connect(models={One, Nine, Seven})
        r = c._resolve_run_order()
        self.assertEqual([{'One', 'Nine'}, {'Seven'}], self.repr_run_order(r.run_order))

    def test_model_iterator(self):
        c = ayeaye.Connect(models={One, Two, Five, Six})
        models = [m for m in c]
        # not ordered when a set of models is passed
        self.assertEqual(4, len(models))

        c = ayeaye.Connect(models=[One, Two, Five, Six])
        models = [m for m in c]
        # ordered when a list
        self.assertEqual([One, Two, Five, Six], models)

    def test_run_order(self):
        """
        run order is either specified when list is passed or resolved by examining dataset
        provenance but that is tested elsewhere.
        Here, check there is a public method that returns a list and elements within this list
        are a tree of lists and sets.
        """
        def is_run_item(r):
            if isinstance(r, list):
                # all items in list must be a set. Set could be one model
                for r2 in r:
                    assert isinstance(r2, set)
                    is_run_item(r2)
            elif isinstance(r, set):
                for r2 in r:
                    assert issubclass(r2, ayeaye.Model) or isinstance(r2, list)
                    if isinstance(r2, list):
                        is_run_item(r2)
            else:
                raise ValueError("Non list and not set item found")
            return True

        for c in [ayeaye.Connect(models={One, Two, Five, Six}),
                  ayeaye.Connect(models=[One, Two, Five, Six])
                  ]:
            run_order = c.run_order()
            self.assertIsInstance(run_order, list)
            self.assertTrue(is_run_item(run_order))



import unittest

from ayeaye.ignition import Ignition, EngineUrlCase, EngineUrlStatus

EXAMPLE_ENGINE_URL_0 = "mysql://root:{env_secret_password}@localhost/my_database"


class TestIgnition(unittest.TestCase):

    def test_get_raw(self):
        """
        raw is always available because that's the url passed to constructor.
        """
        i = Ignition(EXAMPLE_ENGINE_URL_0)
        status, e_url = i.engine_url_at_state(EngineUrlCase.RAW)
        self.assertEqual(EngineUrlStatus.OK, status)
        self.assertEqual(EXAMPLE_ENGINE_URL_0, e_url)



import json
import os
import shutil
import tempfile
import unittest

import ayeaye

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_CSV_PATH = os.path.join(PROJECT_TEST_PATH, 'data', 'deadly_creatures.csv')


class FakeModel(ayeaye.Model):
    animals = ayeaye.Connect(engine_url=f"csv://{EXAMPLE_CSV_PATH}")

    def build(self):
        for a in self.animals:
            self.log(a.common_name)


class TestModels(unittest.TestCase):

    def setUp(self):
        self._working_directory = None

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_go_closes_dataset_connections(self):

        m = FakeModel()
        m.log_to_stdout = False
        m.go()

        self.assertEqual(None, m.animals.file_handle, "File handle should be closed")

    def test_double_usage(self):
        """
        Use Connect as a callable to make a second iterable. Make the cartesian product to
        demonstrate they are independent.
        """
        class AnimalsModel(ayeaye.Model):
            animals_a = ayeaye.Connect(engine_url="csv://" + EXAMPLE_CSV_PATH)
            animals_b = animals_a.clone()
            animals_output = ayeaye.Connect(access=ayeaye.AccessMode.WRITE)

            def build(self):
                cartesian = []
                for a in self.animals_a:
                    for b in self.animals_b:
                        cartesian.append(f"{a.common_name}_{b.common_name}")
                self.animals_output.data = cartesian

        self.assertNotEqual(id(AnimalsModel.animals_a), id(AnimalsModel.animals_b))

        m = AnimalsModel()

        output_file = "{}/animals_summary.json".format(self.working_directory())
        m.animals_output.update(engine_url=f"json://{output_file};indent=4")
        output_encoding = m.animals_output.encoding

        m.go()

        with open(output_file, 'r', encoding=output_encoding) as f:
            output_data = json.load(f)

        expected_data = ['Crown of thorns starfish_Crown of thorns starfish',
                         'Crown of thorns starfish_Golden dart frog'
                         ]
        self.assertEqual(expected_data, output_data)
